##
# Copyright (c) 2013 Yury Selivanov
# License: Apache 2.0
##
"""Greensocket (non-blocking) for asyncio.

Use ``greenio.socket`` in the same way as you would use stdlib's
``socket.socket`` in ``greenio.task`` tasks or coroutines invoked
from them.
"""
from __future__ import absolute_import
from greenio import asyncio
from socket import error, SOCK_STREAM
from socket import socket as std_socket

from . import yield_from
from . import _GreenLoopMixin


class socket:

    def __init__(self, *args, **kwargs):
        _from_sock = kwargs.pop('_from_sock', None)
        if _from_sock:
            own_sock = None
            self._sock = _from_sock
        else:
            own_sock = std_socket(*args, **kwargs)
            self._sock = own_sock
        try:
            self._sock.setblocking(False)
            self._loop = asyncio.get_event_loop()
            assert isinstance(self._loop, _GreenLoopMixin), \
                'greenio event loop is required'
        except:
            if own_sock is not None:
                # An unexpected error has occurred. Close the
                # socket object if it was created in __init__.
                own_sock.close()
            raise

    @classmethod
    def from_socket(cls, sock):
        return cls(_from_sock=sock)

    @property
    def family(self):
        return self._sock.family

    @property
    def type(self):
        return self._sock.type

    @property
    def proto(self):
        return self._sock.proto

    def _proxy(attr):
        def proxy(self, *args, **kwargs):
            meth = getattr(self._sock, attr)
            return meth(*args, **kwargs)

        proxy.__name__ = attr
        proxy.__qualname__ = attr
        proxy.__doc__ = getattr(getattr(std_socket, attr), '__doc__', None)
        return proxy

    def _copydoc(func):
        func.__doc__ = getattr(
            getattr(std_socket, func.__name__), '__doc__', None)
        return func

    @_copydoc
    def setblocking(self, flag):
        if flag:
            raise error('greenio.socket does not support blocking mode')

    @_copydoc
    def recv(self, nbytes):
        fut = self._loop.sock_recv(self._sock, nbytes)
        yield_from(fut)
        return fut.result()

    @_copydoc
    def connect(self, addr):
        fut = self._loop.sock_connect(self._sock, addr)
        yield_from(fut)
        return fut.result()

    @_copydoc
    def sendall(self, data, flags=0):
        assert not flags
        fut = self._loop.sock_sendall(self._sock, data)
        yield_from(fut)
        return fut.result()

    @_copydoc
    def send(self, data, flags=0):
        self.sendall(data, flags)
        return len(data)

    @_copydoc
    def accept(self):
        fut = self._loop.sock_accept(self._sock)
        yield_from(fut)
        sock, addr = fut.result()
        return self.__class__.from_socket(sock), addr

    @_copydoc
    def makefile(self, mode, *args, **kwargs):
        if mode == 'rb':
            return ReadFile(self._loop, self._sock)
        elif mode == 'wb':
            return WriteFile(self._loop, self._sock)
        raise NotImplementedError

    bind = _proxy('bind')
    listen = _proxy('listen')
    getsockname = _proxy('getsockname')
    getpeername = _proxy('getpeername')
    gettimeout = _proxy('gettimeout')
    getsockopt = _proxy('getsockopt')
    setsockopt = _proxy('setsockopt')
    fileno = _proxy('fileno')
    # socket.detach() was added in Python 3.2
    if hasattr(std_socket, 'detach'):
        detach = _proxy('detach')
    close = _proxy('close')
    shutdown = _proxy('shutdown')

    del _copydoc, _proxy


class ReadFile:

    def __init__(self, loop, sock):
        self._loop = loop
        self._sock = sock
        self._buf = bytearray()

    def read(self, size):
        while 1:
            if size <= len(self._buf):
                data = self._buf[:size]
                del self._buf[:size]
                return data

            fut = self._loop.sock_recv(self._sock, size - len(self._buf))
            yield_from(fut)
            res = fut.result()
            self._buf.extend(res)

            if size <= len(self._buf):
                data = self._buf[:size]
                del self._buf[:size]
                return data
            else:
                data = self._buf[:]
                del self._buf[:]
                return data

    def close(self):
        pass


class WriteFile:

    def __init__(self, loop, sock):
        self._loop = loop
        self._sock = sock

    def write(self, data):
        fut = self._loop.sock_sendall(self._sock, data)
        yield_from(fut)
        return fut.result()

    def flush(self):
        pass

    def close(self):
        pass


def create_connection(address, timeout=None):
    loop = asyncio.get_event_loop()
    host, port = address

    rslt = yield_from(
        loop.getaddrinfo(host, port, family=0, type=SOCK_STREAM))

    for res in rslt:
        af, socktype, proto, canonname, sa = res
        sock = None

        try:
            sock = socket(af, socktype, proto)
            sock.connect(sa)
            return sock
        except ConnectionError:
            if sock:
                sock.close()

    raise error('unable to connect to {!r}'.format(address))



##
# Copyright (c) 2013 Yury Selivanov
# License: Apache 2.0
##

"""greenio package allows to compose greenlets and asyncio coroutines."""

__all__ = ['task', 'yield_from']


import greenlet
import sys

try:
    import asyncio
except ImportError:
    asyncio = None
try:
    import trollius
except ImportError:
    trollius = None
    if asyncio is None:
        raise
if asyncio is None:
    asyncio = trollius

def _create_task(coro, loop):
    if loop is None:
        loop = asyncio.get_event_loop()
    if hasattr(loop, 'create_task'):
        return loop.create_task(coro)
    else:
        return GreenTask(coro, loop=loop)


if trollius is not None:
    def _async(future, loop):
        # trollius iscoroutine() accepts trollius and asyncio coroutine
        # objects
        if trollius.iscoroutine(future):
            return _create_task(future, loop)
        else:
            return future
else:
    def _async(future, loop):
        if asyncio.iscoroutine(future):
            return _create_task(future, loop)
        else:
            return future


_FUTURE_CLASSES = (asyncio.Future,)
if trollius is not None and trollius is not asyncio:
    _FUTURE_CLASSES += (trollius.Future,)

import sys


class _LoopGreenlet(greenlet.greenlet):
    """Main greenlet (analog to main thread) for the event-loop.

    It's a policy task to provide event-loop implementation with
    its "run_*" methods executed in _LoopGreenlet context"""


class _TaskGreenlet(greenlet.greenlet):
    """Each task (and its subsequent coroutines) decorated with
    ``@greenio.task`` is executed in this greenlet"""


class _GreenTaskMixin(object):
    def __init__(self, *args, **kwargs):
        self._greenlet = None
        super(_GreenTaskMixin, self).__init__(*args, **kwargs)

    def _step(self, value=None, exc=None):
        if self._greenlet is None:
            # Means that the task is not currently in a suspended greenlet
            # waiting for results for "yield_from"
            ovr = super(_GreenTaskMixin, self)._step
            self._greenlet = _TaskGreenlet(ovr)

            # Store a reference to the current task for "yield_from"
            self._greenlet.task = self

            # Now invoke overloaded "Task._step" in "_TaskGreenlet"
            result = self._greenlet.switch(value, exc)

            # If "result" is "_YIELDED" it means that the "yield_from"
            # method was called
            if result is not _YIELDED:
                # And if not - then task jumped out of greenlet without
                # calling "yield_from"
                self._greenlet.task = None
                self._greenlet = None
            else:
                self.__class__._current_tasks.pop(self._loop)
        else:
            # The task is in the greenlet, that means that we have a result
            # for the "yield_from"

            self.__class__._current_tasks[self._loop] = self

            if exc is not None:
                if hasattr(exc, '__traceback__'):
                    tb = exc.__traceback__
                else:
                    tb = sys.exc_info()[2]
                result = self._greenlet.throw(
                    type(exc), exc, tb)
            else:
                result = self._greenlet.switch(value)

            # Again, if "result" is "_YIELDED" then we just called "yield_from"
            # again
            if result is not _YIELDED:
                self._greenlet.task = None
                self._greenlet = None
            else:
                self.__class__._current_tasks.pop(self._loop)


class _GreenLoopMixin(object):
    def _green_run(self, method, args, kwargs):
        return _LoopGreenlet(method).switch(*args, **kwargs)

    def run_until_complete(self, *args, **kwargs):
        ovr = super(_GreenLoopMixin, self).run_until_complete
        return self._green_run(ovr, args, kwargs)

    def run_forever(self, *args, **kwargs):
        ovr = super(_GreenLoopMixin, self).run_forever
        return self._green_run(ovr, args, kwargs)


class GreenTask(_GreenTaskMixin, asyncio.Task):
    pass


class GreenUnixSelectorLoop(_GreenLoopMixin, asyncio.SelectorEventLoop):
    def create_task(self, coro):
        return GreenTask(coro)


class GreenEventLoopPolicy(asyncio.DefaultEventLoopPolicy):
    def new_event_loop(self):
        return GreenUnixSelectorLoop()


if trollius is not None:
    if trollius is not asyncio:
        class GreenTrolliusTask(_GreenTaskMixin, trollius.Task):
            pass

        class GreenTrolliusUnixSelectorLoop(_GreenLoopMixin,
                                            trollius.SelectorEventLoop):
            def create_task(self, coro):
                return GreenTrolliusTask(coro)

        class GreenTrolliusEventLoopPolicy(asyncio.DefaultEventLoopPolicy):

            def new_event_loop(self):
                return GreenTrolliusUnixSelectorLoop()
    else:
        GreenTrolliusTask = GreenTask
        GreenTrolliusUnixSelectorLoop = GreenUnixSelectorLoop
        GreenTrolliusEventLoopPolicy = GreenEventLoopPolicy


def yield_from(future, loop=None):
    """A function to use instead of ``yield from`` statement."""

    future = _async(future, loop)

    gl = greenlet.getcurrent()

    if __debug__:
        if not isinstance(gl.parent, _LoopGreenlet):
            raise RuntimeError(
                '"greenio.yield_from" requires GreenEventLoopPolicy '
                'or compatible')
            # or something went horribly wrong...

        if not isinstance(gl, _TaskGreenlet):
            raise RuntimeError(
                '"greenio.yield_from" was supposed to be called from a '
                '"greenio.task" or a subsequent coroutine')
            # ...ditto

    task = gl.task

    if not isinstance(future, _FUTURE_CLASSES):
        raise RuntimeError(
            'greenlet.yield_from was supposed to receive only Futures, '
            'got {!r} in task {!r}'.format(future, task))

    # "_wakeup" will call the "_step" method (which we overloaded in
    # GreenTask, and therefore wakeup the awaiting greenlet)
    future.add_done_callback(task._wakeup)
    task._fut_waiter = future

    # task cancellation has been delayed.
    if task._must_cancel:
        if task._fut_waiter.cancel():
            task._must_cancel = False

    # Jump out of the current task greenlet (we'll return to GreenTask._step)
    return gl.parent.switch(_YIELDED)


def task(func, loop=None):
    """A decorator, allows use of ``yield_from`` in the decorated or
    subsequent coroutines."""

    coro_func = asyncio.coroutine(func)

    def task_wrapper(*args, **kwds):
        coro_obj = coro_func(*args, **kwds)
        return _create_task(coro_obj, loop)

    return task_wrapper


class _YIELDED(object):
    """Marker, don't use it"""



from setuptools import setup


extra = {}
f = open('README.rst', 'r')
try:
    extra['long_description'] = f.read()
finally:
    f.close()


setup(
    name='greenio',
    version='0.6.0',
    description="Greenlets for asyncio (PEP 3156).",
    url='https://github.com/1st1/greenio/',
    license='Apache 2.0',
    packages=['greenio'],
    install_requires=['greenlet'],
)
